from munch import Munch
from functools import partial

from infi.shell import commandgroup, command
from infi.shell.utils import strlist
from infi.shell.exceptions import BadArguments, BadValue
from infi.shell.types import Str

from .types import Name, Host, Cluster, Volume, MasterVolume, VolumeSnapshotNotWP, Lun, ISCSISecurityMethod
from .options import QueryOptions
from .host import HostCommandsBase
from . import render_collection, roles, Null

from ..features import feature_command


QUERY_FIELDS = strlist("name hosts luns created_at")
DETAILED_QUERY_FIELDS = strlist("name hosts luns created_at updated_at")
SORTABLE_FIELDS = strlist("name created_at updated_at")

HOST_QUERY_FIELDS = strlist("name host")
MAP_QUERY_FIELDS = strlist("cluster volume lun")

@commandgroup(prefix="cluster", provides=(Cluster,))
class ClusterCommands(HostCommandsBase):
    ENTITY_TYPE = Cluster

    @command(
        description="Create a new cluster",
        types=(Name,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name):
        return self._create(_api, name=name)

    @command(
        description="Rename cluster",
        types=(Cluster, Name),
        hidden_substitue=dict(name="cluster"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def rename(self, _api, cluster, new_name):
        return self._rename(_api, cluster, new_name)

    @command(
        description="Delete cluster",
        types=(Cluster,),
        hidden_substitue=dict(name="cluster"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, cluster):
        return self._delete(_api, cluster)

    @command(
        description="Add host to cluster",
        types=(Cluster, Host.List),
        hidden_substitue=dict(name="cluster"),
        substitute=dict(
            hosts="host",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def add_host(self, _api, cluster, hosts):
        with _api.aggregation() as aggregator:
            for host in hosts:
                aggregator.next_task(host.name, "Added")
                self._post(_api, cluster, "hosts", data=dict(id=host.id))
        if len(hosts) == 1:
            return 'Host "{}" added to cluster "{}"'.format(hosts[0].name, cluster.name)
        return '{} hosts added to cluster "{}"'.format(len(hosts), cluster.name)

    @command(
        description="Remove host from cluster",
        types=(Cluster, Host.List),
        hidden_substitue=dict(name="cluster"),
        substitute=dict(
            hosts="host",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove_host(self, _api, cluster, hosts):
        with _api.aggregation() as aggregator:
            for host in hosts:
                aggregator.next_task(host.name, "Removed")
                self._del(_api, cluster, "hosts", host.id)
        if len(hosts) == 1:
            return 'Host "{}" removed from cluster "{}"'.format(host.name, cluster.name)
        return '{} hosts removed from cluster "{}"'.format(len(hosts), cluster.name)

    @command(
        description="Map {} to a cluster".format(Volume.DESCRIPTION_SUBJECT),
        types=(Cluster, MasterVolume.List, VolumeSnapshotNotWP.List, Lun),
        hidden_substitue=dict(name="cluster"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def map(self, _api, cluster, vol=None, snap=None, lun=None):
        return self._map(_api, cluster, lun, vol, snap)

    @command(
        description="Unmap {} from cluster".format(Volume.DESCRIPTION_SUBJECT),
        types=(Cluster, MasterVolume.List, VolumeSnapshotNotWP.List, Lun),
        hidden_substitue=dict(name="cluster"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def unmap(self, _api, cluster, vol=None, snap=None, lun=None):
        return self._unmap(_api, cluster, lun, vol, snap)

    @command(
        description="List existing clusters",
        types=(Cluster.List,),
        hidden_substitue=dict(name="cluster"),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, detailed=True),
        )
    @render_collection(
        entity=Cluster,
        fields=QUERY_FIELDS,
        alignment=dict(
            hosts="right",
            ),
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, cluster=None):
        return self._query(_api, cluster, _hook=self._query_hook)

    def _query_hook(self, clusters):
        for cluster in clusters:
            cluster.hosts = len(cluster.hosts)
            cluster.luns = len(cluster.luns)
        return clusters

    @command(
        description="List existing clusters with member hosts",
        types=(Cluster.List, Host.List),
        hidden_substitue=dict(name="cluster"),
        options=QueryOptions(fields=HOST_QUERY_FIELDS, sortable=HOST_QUERY_FIELDS, detailed=False),
        )
    @render_collection(
        entity=Cluster,
        )
    def host_query(self, _api, cluster=None, host=None):
        if cluster and host:
            raise BadArguments("Either cluster, host, or neither may be specified")
        host_names = set(h.name for h in host) if host else None
        return self._query(_api,
            cluster,
            _hook=partial(self._host_query_hook, host_names=host_names),
            _sort_on_client_side=True,
            )

    def _host_query_hook(self, clusters, host_names):
        return [Munch(
            name=cluster.name,
            host=host.name
            )
            for cluster in clusters
            for host in sorted(cluster.hosts, key=lambda h: h.name)
            if not host_names or host.name in host_names]

    @command(
        description="List existing cluster mappings",
        types=(Cluster,),
        hidden_substitue=dict(name="cluster"),
        options=QueryOptions(fields=MAP_QUERY_FIELDS),
        )
    @render_collection(
        entity=Cluster,
        fields=MAP_QUERY_FIELDS,
        substitute=dict(
            lun="lun_id",
            ),
        )
    def map_query(self, _api, cluster=None):
        return self._map_query(_api, cluster)

    @feature_command(
        feature="iscsi",
        status=True,
        description="Set iSCSI authentication method for all hosts in cluster",
        types=(Cluster, ISCSISecurityMethod),
        hidden_substitue=dict(name="cluster"),
        substitute=dict(
            method="type",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_auth_type(self, _api, cluster, method):
        if not cluster.hosts:
            raise BadValue(self._result_message(cluster, "has no hosts"))
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for host in cluster.hosts:
                task_name = self._result_subject_and_entity(host, entity_type=Host, capitalize=True)
                props, message = self._prepare_host_iscsi_security_props(host, method)
                if not props:
                    aggregator.next_task(task_name, message)
                    aggregator.mark_current_task_successful()
                    continue
                aggregator.next_task(task_name, message)
                self._update(_api, host, _entity_type=Host, returns=None, **props)
        return aggregator.run()

    @feature_command(
        feature="iscsi",
        status=True,
        description="Set iSCSI CHAP usernames and secrets for all hosts in cluster",
        types=(Cluster, Str, Str, Str, Str),
        hidden_substitue=dict(name="cluster"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_chap(self, _api, cluster, inbound_user=None, inbound_secret=None, outbound_user=None, outbound_secret=None):
        if not cluster.hosts:
            raise BadValue(self._result_message(cluster, "has no hosts"))
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._result_message(cluster, "CHAP settings updated"),
            )
        config = _api.get('config')['iscsi']
        with aggregator:
            for host in cluster.hosts:
                outbound_user = outbound_user if outbound_user or host.security_chap_outbound_username else config['iscsi.default_iqn']
                inbound_user = inbound_user if inbound_user or host.security_chap_inbound_username else host.ports[0].address if host.ports else None
                task = aggregator.next_task(self._result_subject_and_entity(host, entity_type=Host, capitalize=True), "CHAP settings updated")
                self._set_host_chap_settings(_api, host, inbound_user, inbound_secret, outbound_user, outbound_secret, aggregator_task=task)
        return aggregator.run()


    @feature_command(
        feature="iscsi",
        status=True,
        description="Clear CHAP usernames and secrets for all hosts in cluster",
        types=(Cluster,),
        hidden_substitue=dict(name="cluster"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear_chap(self, _api, cluster):
        if not cluster.hosts:
            raise BadValue(self._result_message(cluster, "has no hosts"))
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._result_message(cluster, "CHAP settings cleared"),
            )

        with aggregator:
            for host in cluster.hosts:
                task = aggregator.next_task(self._result_subject_and_entity(host, entity_type=Host, capitalize=True), "CHAP settings cleared")
                self._update(_api, host, _entity_type=Host, _multi_field_put=True, returns=None,
                    security_method="NONE",
                    security_chap_inbound_username=Null,
                    security_chap_inbound_secret=Null,
                    security_chap_outbound_username=Null,
                    security_chap_outbound_secret=Null,
                    )
        return aggregator.run()
