from collections import OrderedDict
import sys
import os

from infi.shell.commands import commandgroup, command, render
from infi.shell.types import (
    IPAddress, Hostname, YesNo, Str
    )
from infi.shell.exceptions import BadArguments, BadValue

from ....environment import variables as env
from ..types import Name, EthPort, SessionTimeout
from ..utils import increased_api_timeout
from .. import roles
from . import ConfigCommandsBase

@commandgroup(parent=ConfigCommandsBase, prefix="system")
class SystemConfigCommands(ConfigCommandsBase):

    def _format_ips_string(self, s):
        return (s or "").replace(",", ", ")

    @command(
        description="Display general system configuration",
        )
    @render(
        individually=True,
        )
    def query(self, _api):
        config = self._get_config(_api)
        result = OrderedDict((
            ("name", _api.get("system/name")),
            ("management_ip_address", "{} / {}".format(config.ip_config.ip, config.ip_config.netmask)),
            ("default_gateway", config.ip_config.gateway),
            ("ssl_Redirection", "{}".format("enabled" if config.mgmt['network.force_ssl_all'] else "disabled")),
            ("active_session_timeout", config.mgmt['mgmt.session_expiry']),
            ("idle_session_timeout", config.mgmt['mgmt.session_idle_timeout']),
            ("default_pool_compression", config.mgmt['pool.compression_enabled_default']),
            ("default_pool_ssd_cache", config.mgmt['pool.ssd_enabled_default']),
            ))
        node_ips = []
        for port in EthPort.iter_all(_api):
            if port.role and port.role.lower() == "management":
                node_ips.append((port.node_index, port.ip_v4_addr))
        result.update(("node_{}_ip_address".format(node_index), ip) for node_index, ip in sorted(node_ips))
        result.update((
            ("dns_servers", self._format_ips_string(config.mgmt["environment.dns_servers"])),
            ("ntp_servers", self._format_ips_string(config.mgmt["environment.ntp_servers"])),
            ))
        motd = config.mgmt['client.message_of_the_day']
        result.update((
            ("message_of_the_day", u"{}".format(motd)),
            ))
        return result;

    @command(
        description="Set system timeouts",
        types=(SessionTimeout, SessionTimeout),
        )
    @roles.INFINIDAT.ADMIN
    def set_session_timeout(self, _api, active_session=None, idle_session=None):
        if active_session is None and idle_session is None:
            raise BadArguments, "Atleast one of the parameters must be provided"
        
        current_idle_session = int(self._get_param(_api, "mgmt", "mgmt.session_idle_timeout"))
        current_active_session = int(self._get_param(_api, "mgmt", "mgmt.session_expiry"))

        actions = dict(
            idle=("mgmt.session_idle_timeout", idle_session, "Session idle timeout", "{} minutes".format(idle_session)),
            active=("mgmt.session_expiry", active_session, "Session expiry", "{} minutes".format(active_session))
            )

        order = ['idle', 'active']                
        if idle_session is None:
            actions.pop('idle', None)
            if active_session < current_idle_session:
                raise BadArguments, "Active session timeout must be beggier the idle session timeout ({})".format(current_idle_session)
        elif active_session is None:
            actions.pop('active', None)
            if idle_session > current_active_session:
                raise BadArguments, "Idle session timeout must be smaller then active session timeout ({})".format(current_active_session)
        else:
            if active_session < idle_session:
                raise BadArguments, "Given active session timeout must be bigger the given idle session timeout ({})".format(idle_session)

        if active_session >= current_active_session:
            order = ['active', 'idle']
            
        answers = []
        for i in order:
            command = actions.pop(i, None)
            if command: 
                answers.append(self._configure_param(_api,
                    category="mgmt",
                    param=command[0],
                    value=command[1],
                    subject=command[2],
                    value_name=command[3],
                    )
                )
        answers = '. '.join(answers)
        return answers

    @command(
        description="Set DNS server IPs",
        types=(IPAddress, IPAddress),
        )
    @roles.INFINIDAT.ADMIN
    def set_dns(self, _api, primary, secondary=None):
        servers = [primary.ip]
        if secondary:
            servers.append(secondary.ip)
        return self._configure_param(_api,
            category="mgmt",
            param="environment.dns_servers",
            value=",".join(servers),
            subject="DNS server" + ("s" if len(servers) > 1 else ""),
            value_name=", ".join(servers),
            )

    @command(
        description="Set system name",
        types=(Name,),
        )
    @roles.INFINIDAT.ADMIN
    def set_name(self, _api, name):
        current_name = _api.get("system/name")
        if current_name == name:
            return 'System name is already "{}"'.format(name)
        _api.put("system/name", name)
        return 'System renamed from "{}" to "{}"'.format(current_name, name)

    @command(
        description="Enables / disables the redirection from HTTP (80)to HTTPS (443)",
        types=(YesNo,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def set_ssl_redirection(self, _api, redirect):
        system = _api.get(self._BASE_URI)
        if system.mgmt["network.force_ssl_all"] == redirect :
            return "HTTPS redirection is already {}".format("enabled" if redirect else "disabled")
        return self._configure_param(_api,
            category="mgmt",
            param="network.force_ssl_all",
            value=redirect,
            subject="HTTPS redirection",
            value_name="{}".format("enabled" if redirect else "disabled"),
        )

    @command(
        description="Set message of the day",
        types=(YesNo, ),
        )
    @roles.INFINIDAT.ADMIN
    def set_motd(self, _api, clear=None):
        motd = ''
        if not clear:
            print u'Please enter message of the day below.'
            print u'Use ctrl+D to finish:\n'
            buffer = []
            while True:
                userinput = sys.stdin.readline()
                if not userinput:
                    break
                userinput = userinput.decode('utf-8')
                buffer.append(userinput)

            motd = u''.join(buffer)

        return self._configure_param(_api,
            category="mgmt",
            param="client.message_of_the_day",
            value=motd,
            subject="Message of the day",
            value_name=motd,
            )

    @command(
        description="Enables / disables default pool compression",
        types=(YesNo,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def set_default_pool_compression(self, _api, compression):
        system = _api.get(self._BASE_URI)
        if system.mgmt["pool.compression_enabled_default"] == compression :
            return "Default pool compression is already {}".format("enabled" if compression else "disabled")
        return self._configure_param(_api,
            category="mgmt",
            param="pool.compression_enabled_default",
            value=compression,
            subject="Default pool compression",
            value_name="{}".format("enabled" if compression else "disabled"),
        )

    @command(
        description="Enables / disables default pool ssd cache",
        types=(YesNo,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def set_default_pool_ssd_cache(self, _api, ssd_cache):
        system = _api.get(self._BASE_URI)
        if system.mgmt["pool.ssd_enabled_default"] == ssd_cache :
            return "Default pool compression is already {}".format("enabled" if ssd_cache else "disabled")
        return self._configure_param(_api,
            category="mgmt",
            param="pool.ssd_enabled_default",
            value=ssd_cache,
            subject="Default pool ssd cache",
            value_name="{}".format("enabled" if ssd_cache else "disabled"),
        )

@commandgroup(parent=SystemConfigCommands, prefix="time")
class SystemTimeCommands(ConfigCommandsBase):

    @command(
        description="Set NTP servers",
        types=(Hostname, Hostname),
        )
    @roles.INFINIDAT.ADMIN
    def set_ntp(self, _api, primary, secondary=None):
        servers = [primary]
        if secondary:
            servers.append(secondary)
        return self._configure_param(_api,
            category="mgmt",
            param="environment.ntp_servers",
            value=",".join(servers),
            subject="NTP server" + ("s" if len(servers) > 1 else ""),
            value_name=", ".join(servers),
            )


@commandgroup(parent=SystemConfigCommands, prefix="ssl_certificate")
class SystemSSLCommands(ConfigCommandsBase):

    @command(
        description="Sets the SSL certificate for the system",
        types=(Str,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def upload(self, _api, filename):
        if os.environ.get(env._prefix + '_AUTH_JSESSIONID'):
            return "Uploading a certificate is available from either the GUI or stand-alone CLI"
        if not os.path.isfile(filename):
            raise BadValue("File {} was not found".format(filename))
        if os.path.getsize(filename) > 1024 * 1024:
            raise BadValue("Allowed package size is at most 1 MiB")
        with increased_api_timeout():
            _api.upload("system/certificates", (filename,))
        return "System SSL certificate updated successfully"

    @command(
        description="Clears existing SSL certificate, replaces it with a self-signed certificate",
        types=(),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def clear(self, _api):
        _api.post("system/certificates/generate_self_signed")
        return " SSL certificate cleared, generated a self-signed certificate."

    @command(
        description="Create Certificate Sign Request (CSR) file. This file includes only the public key and is a more secure method of generating a certificate.",
        types=(Str,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def create_csr(self, _api, filepath=""):
        if os.environ.get(env._prefix + '_AUTH_JSESSIONID'):
            return "Downloading a certificate is available from either the GUI or stand-alone CLI"
        if filepath and not os.path.isdir(filepath):
            raise BadValue("Path {} was not found".format(filepath))
        with increased_api_timeout():
            result = _api.post("system/certificates/generate_csr")
        sep = '\\' if os.name == "nt" else '/'
        sep = sep if (filepath[-1] if filepath else "") not in ['/','\\'] else ""
        filepath = "{}infinibox_csr.pem".format(filepath + sep if filepath else "")
        try:
            f = open(filepath, "w")
            f.write(result)
            f.close()
        except Exception as e:
            raise BadValue, "{}, {}".format(e.errno, e.strerror)

        return "CSR saved to '{}'".format(filepath)

    @command(
        description="Sets the certificate using a signed CSR file.",
        types=(Str,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def upload_signed_csr(self, _api, filename):
        if os.environ.get(env._prefix + '_AUTH_JSESSIONID'):
            return "Uploading a certificate is available from either the GUI or stand-alone CLI"
        if not os.path.isfile(filename):
            raise BadValue("File {} was not found".format(filename))
        if os.path.getsize(filename) > 1024 * 1024:
            raise BadValue("Allowed package size is at most 1 MiB")
        with increased_api_timeout():
            _api.upload("system/certificates/upload_csr", (filename,))
        return "Signed CSR uploaded successfully"

    @command(
        description="Show current certificate data",
        types=(),
        )
    @render(
        individually=True,
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def query(self, _api):
        result = _api.get("system/certificates")
        def flatten_dict(to_flat, top_key='', sep='_'):
            res = []
            for k, v in to_flat.items():
                key = top_key + sep + k if top_key else k
                if isinstance(v, dict):
                    res.extend(flatten_dict(v, key).items())
                else:
                    res.append((key, v))
            return OrderedDict(res)
        result = flatten_dict(result[0])
        result["public_key_modulus"] = "\n".join([result["public_key_modulus"][i:i+64] for i in range(0, len(result["public_key_modulus"]), 64)])
        values_to_replace = OrderedDict([
            ('certificate_issued_by_C', 'certificate_issued_by_country'),
            ('certificate_issued_by_O', 'certificate_issued_by_organization'),
            ('certificate_issued_by_CN', 'certificate_issued_by_cn'),
            ('certificate_issued_to_C', 'certificate_issued_to_country'),
            ('certificate_issued_to_O', 'certificate_issued_to_organization'),
            ('certificate_issued_to_CN', 'certificate_issued_to_cn')
            ])
        for old, new in values_to_replace.items():
            result[new] = result.pop(old)
        return result
        